/***********************************************************************
 * Time-stamp: <Thursday, 19.07.2001 at 19:13:33; edited by bauke>
 * 
 * Tina's random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#include <iostream>
#include <cstdlib>
#include <new>
#include <cmath>
#include <vector>
#include "trng.h"

using namespace std;
using namespace TRNG;


const char * TRNG::version(void) {
  return("TRNG V1.0.1");
}


long TRNG::modulo_invers(long a, long m) {
  // a modulus invers for prime modulus m
  // 1/a mod m = a^(m-2)
  long long p, al;
  long e;
  if (a==0)
    return(0);
  else {
    al=static_cast<long long>(a);
    e=m-2;
    p=1;
    while (e>0) {
      if ((e&1)==1) {
	p*=al;
	p%=m;
      }
      al*=al;
      al%=m;
      e>>=1;
    }
    return(static_cast<long>(p));
  }
}


void TRNG::gauss(vector<long> &a, vector<long> &b, long m) {
  // Gauss algorithm in modular aithmetic
  // a[0..n^2-1] : matrix
  // b[0..n-1]   : vector
  // m           : modulus
  // returns in b the solution
  long i, j, k, t, n;
  // initialize indices
  n=b.size();
  vector<long> p(n);
  for (i=0; i<n; ++i)
    p[i]=i;
  // make matrix diagonal
  for (i=0; i<n-1; ++i) {
    if (a[n*p[i]+i]==0) {
      // swap
      j=i+1;
      while (j<n && a[n*p[j]+i]==0) 
	j++;
      if (j==n)
	throw error("singular matrix in TRNG::gauss");
      t=p[i];  p[i]=p[j];  p[j]=t;
    }
    t=modulo_invers(a[n*p[i]+i], m);
    for (j=i; j<n; j++) 
      a[n*p[i]+j]=static_cast<long>
	((static_cast<long long>(a[n*p[i]+j])*
	  static_cast<long long>(t))%m);
    b[p[i]]=static_cast<long>
      ((static_cast<long long>(b[p[i]])*
	static_cast<long long>(t))%m);
    for (j=i+1; j<n; j++) {
      if (a[n*p[j]+i]!=0) {
	t=modulo_invers(a[n*p[j]+i], m);
	for (k=i; k<n; k++) {
	  a[n*p[j]+k]=
	    static_cast<long>
	    ((static_cast<long long>(a[n*p[j]+k])*
	      static_cast<long long>(t))%m);
	  a[n*p[j]+k]-=a[n*p[i]+k];
	  if (a[n*p[j]+k]<0)
	    a[n*p[j]+k]+=m;
	}
	b[p[j]]=static_cast<long>
	  ((static_cast<long long>(b[p[j]])*
	    static_cast<long long>(t))%m);
	b[p[j]]-=b[p[i]];
	if (b[p[j]]<0)
	  b[p[j]]+=m;
      }
    }
  }
  // solve diagonal system
  t=modulo_invers(a[n*p[n-1]+n-1], m);
  a[n*p[n-1]+n-1]=1;
  b[p[n-1]]=static_cast<long>
    ((static_cast<long long>(b[p[n-1]])*
      static_cast<long long>(t))%m);
  for (i=n-2; i>=0; i--)
    for (j=i+1; j<n; j++) {
      b[p[i]]-=static_cast<long>
	((static_cast<long long>(a[n*p[i]+j])*
	  static_cast<long long>(b[p[j]]))%m);
      if (b[p[i]]<0)
	b[p[i]]+=m;
    }
  // sort
  for (i=0; i<n; ++i) 
    p[i]=b[p[i]];
  for (i=0; i<n; ++i) 
    b[i]=p[i];
}


void TRNG::matrix_mult(vector<long> &a, vector<long> &b, 
		       vector<long> &c, long m) {
  // multiply two matrices a and b
  // c=a*b
  // n : dimension of a and b
  // m : modulus
  long i, j, k, n;
  long long t;
  n=a.size()<b.size() ? a.size() : b.size();
  if (a.size()!=b.size())
    throw error("different sized vectors in TRNG::matrix_mult");
  n=static_cast<long>(sqrt(static_cast<double>(n)));
  for (i=0; i<n; ++i)
    for (j=0; j<n; j++) {
      t=0;
      for (k=0; k<n; k++) {
	t+=(static_cast<long long>(a[j*n+k])*
	    static_cast<long long>(b[k*n+i]))%m;
	if (t>=m)
	  t-=m;
      }
      c[j*n+i]=static_cast<long>(t);
    }
}


void TRNG::matrix_vec_mult(vector<long> &a, vector<long> &b, 
			   vector<long> &c, long m) {
  // multiply a matrix a and a vector b
  // c=a*b
  // n : dimension of a and b
  // m : modulus
  long j, k, n;
  long long t;
  n=b.size();
  if (a.size()!=b.size()*b.size())
    throw error("different sized vectors in TRNG::matrix_vec_mult");
  for (j=0; j<n; j++) {
    t=0;
    for (k=0; k<n; k++) {
      t+=(static_cast<long long>(a[j*n+k])*
	  static_cast<long long>(b[k]))%m;
      if (t>=m)
	t-=m;
    }
    c[j]=static_cast<long>(t);
  }
}


long long TRNG::scalar_prod(vector<long long> &a, vector<long long> &b) {
  // returns scalarproduct of two vectors a and b with dimension n
  long long q;
  long i, n;
  n=a.size()<b.size() ? a.size() : b.size();
  if (a.size()!=b.size())
    throw error("different sized vectors in TRNG::scalar_prod");
  q=0ll;
  for (i=0; i<n; ++i) {
    q+=a[i]*b[i];
  }
  return(q);
}


//----------------------------------------------------------------------


const char * TRNG::RNG::name(void) {
  return("RNG");
}

void TRNG::RNG::reset(void) {
}

void TRNG::RNG::seed(long s=0l) {
}

long TRNG::RNG::rand(void) {
  return(0l);
}

long TRNG::RNG::max(void) {
  return(max_val);
}

bool TRNG::RNG::boolean(void) {
  return((rand()<max_val2) ? true : false);
}

bool TRNG::RNG::boolean(double p) {
  return((rand()<static_cast<long>(p*max())+1.0) ? true : false);
}

double TRNG::RNG::uniform(void) {
  return(uniformco());
}

double TRNG::RNG::uniformco(void) {
  return(static_cast<double>(rand())/
	 (static_cast<double>(max())+1.0));
}

double TRNG::RNG::uniformcc(void) {
  return(static_cast<double>(rand())/
	 static_cast<double>(max()));
}

double TRNG::RNG::uniformoo(void) {
  return((static_cast<double>(rand())+1.0)/
	 (static_cast<double>(max())+2.0));
}

double TRNG::RNG::uniformoc(void) {
  return((static_cast<double>(rand())+1.0)/
	 (static_cast<double>(max())+1.0));
}

double TRNG::RNG::exp_dist(double mu) {
  double t;
  if (mu<=0.0)
    throw error("negative or zero parameter in TRNG::RNG::exp_dist");
  t=uniformoc();
  return(-mu*log(t));
}

double TRNG::RNG::normal_dist(double sigma, double mean) {
  static double s, t1, t2;
  static bool calc=true;
  if (sigma<=0.0)
    throw error("negative or zero standard deviation TRNG::RNG::normal_dist");
  if (calc) {
    calc=false;
    do {
      t1=2.0*uniformco()-1.0;
      t2=2.0*uniformco()-1.0;
      s=t1*t1+t2*t2;
    } while (s>=1.0);
    s=sqrt(-2.0*log(s)/s);
    return(t1*s*sigma+mean);
  } else {
    calc=true;
    return(t2*s*sigma+mean);
  }
}

double TRNG::RNG::Gamma_dist(double p1) {
  const double e=2.71828182845904524;
  const double pi=3.1415926535897932;
  double t1, t2, v, x, y, u1, u2, v1, v2;
  if (p1<=0.0)
    throw error("parameter less or equal 0.0 in TRNG::RNG::Gamma_dist");
  if (p1<1.0) {
    v1=(e+p1)/e;
    while (true) {
      u1=uniformco();
      u2=uniformco();
      v2=v1*u1;
      if (v2<=1.0) {
	x=pow(v2, 1.0/p1);
	if (u2<=exp(-x))
	  return(x);
      } else {
	x=-log((v1-v2)/p1);
	if (u2<=pow(x, p1-1.0))
	  return(x);
      }
    }
  }
  if (p1>1.0) {
    t1=sqrt(2.0*p1-1.0);
    t2=p1-1.0;
    do {
      do {
	y=tan(pi*uniformco());
	x=t1*y+t2;
      } while (x<=0.0);
      v=uniformco();
    } while (v>(1.0+y*y)*exp(t2*log(x/t2)-t1*y));
    return(x);
  }
  return(exp_dist(1.0));
}

double TRNG::RNG::Beta_dist(double a, double b) {
  double t1, t2;
  if (a<=0.0 || b<=0.0)
    throw error("negative or zero parameter in TRNG::RNG::Beta_dist");
  if (a>1.0 || b>1.0) {
    t1=Gamma_dist(a);
    t2=Gamma_dist(b);
  } else {
    a=1.0/a;
    b=1.0/b;
    do {
      t1=pow(uniformco(), a);
      t2=pow(uniformco(), b);
    } while (t1+t2>1.0);
  }
  return(t1/(t1+t2));
}

double TRNG::RNG::chi_square_dist(double nu) {
  if (nu<1.0)
    throw error("parameter less than one in TRNG::RNG::chi_square_dist");
  return(2.0*Gamma_dist(0.5*nu));
}

double TRNG::RNG::Student_t_dist(double nu) {
  if (nu<=0.0)
    throw error("parameter less than zero in TRNG::RNG::Student_t_dist");
  return(normal_dist(1.0, 0.0)*sqrt(2.0*nu/Gamma_dist(0.5*nu)));
}

long TRNG::RNG::binomial_dist(long t, double p) {
  long k, j;
  if (p<=0.0 || p>=1.0)
    throw error("probability <=0.0 or >=1.0 in TRNG::RNG::binomial_dista");
  if (t<=0l)
    throw error("less than one trail TRNG::RNG::binomial_dista");
  k=0l;
  for (j=0l; j<t; ++j)
    if (boolean(p))
      ++k;
  return(k);
}

long TRNG::RNG::poisson_dist(double mu) {
  long k;
  double A;
  if (mu<=0)
    throw error("parameter less than 0 in TRNG::RNG::poisson_dista");
  if (mu<10l) {
    mu=exp(-mu);
    k=-1l;
    A=1.0;
    do {
      ++k;
      A*=uniformco();
    } while (A>mu);
    return(k);
  } else {
    k=static_cast<long>(rint(normal_dist(sqrt(mu), mu)));
    return(k>=0l ? k : 0l);
  }
}

double TRNG::RNG::rejection(double (*f)(double), double a1, double a2, 
			    double p_max) {
  double t1, t2;
  do {
    t1=uniformco()*(a2-a1)+a1;
    t2=uniformco()*p_max;
  } while ((*f)(t1)<t2);
  return(t1);
}

vector2d TRNG::RNG::spherical2d(void) {
  double r2, t1, t2, t12, t22;
  vector2d vec;
  do {
    t1=2.0*uniformco()-1.0;
    t2=uniformco();
    t12=t1*t1;
    t22=t2*t2;
    r2=t12+t22;
  } while (r2>1.0);
  vec.x1=2.0*t1*t2/r2;
  vec.x2=(t12-t22)/r2;
  return(vec);
}

vector3d TRNG::RNG::spherical3d(void) {
  double q, r2, t1, t2;
  vector3d vec;
  do {
    t1=2.0*uniformco()-1.0;
    t2=2.0*uniformco()-1.0;
    r2=t1*t1+t2*t2;
  } while (r2>1.0);
  q=2.0*sqrt(1.0-r2);
  vec.x1=t1*q;
  vec.x2=t2*q;
  vec.x3=1-2*r2;
  return(vec);
}

vector4d TRNG::RNG::spherical4d(void) {
  double q, r21, r22, t1, t2, t3, t4;
  vector4d vec;
  do {
    t1=2.0*uniformco()-1.0;
    t2=2.0*uniformco()-1.0;
    r21=t1*t1+t2*t2;
  } while (r21>1.0);
  do {
    t3=2.0*uniformco()-1.0;
    t4=2.0*uniformco()-1.0;
    r22=t3*t3+t4*t4;
  } while (r22>1.0);
  q=sqrt((1.0-r21)/r22);
  vec.x1=t1;
  vec.x2=t2;
  vec.x3=t3*q;
  vec.x4=t4*q;
  return(vec);
}

RNG *TRNG::RNG::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new RNG);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::RNG::split(long s, long n) {
  throw error("TRNG::RNG::split not implemented");
}

void TRNG::RNG::jump(long s) {
  throw error("TRNG::RNG::jump not implemented");
}

void TRNG::RNG::save_status(vector<long> &s) {
  throw error("TRNG::RNG::save_status not implemented");
}

void TRNG::RNG::load_status(vector<long> &s) {
  throw error("TRNG::RNG::load_status not implemented");
}

TRNG::RNG::RNG() {
}


//----------------------------------------------------------------------


const char * TRNG::generic_MLCG::name(void) {
  return("generic_MLCG");
}

void TRNG::generic_MLCG::backward(void) {
  r=static_cast<long>
    ((static_cast<long long>(TRNG::modulo_invers(a, modulus))*
      static_cast<long long>(r))%modulus);
}

void TRNG::generic_MLCG::reset(void) {
  a=a_save;
  r=r_save;
  modulus=modulus_save;
  max_val=modulus-1l;;
  max_val2=max_val/2l;
}

void TRNG::generic_MLCG::seed(long s=0l) {
  s++;
  if ((s%modulus)<=0l)
    throw error("bad seed for TRNG::generic_MLCG::seed");
  r=s%modulus;
}

long TRNG::generic_MLCG::rand(void) {
  r=static_cast<long>
    ((static_cast<long long>(a)*static_cast<long long>(r))%modulus);
  return(r);
}

RNG *TRNG::generic_MLCG::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new generic_MLCG(1l, 1l, 1l));
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::generic_MLCG::split(long s, long n) {
  long b, c, i;
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::generic_MLCG::split");
  if (s>1l) {
    b=1l;
    c=a;
    for (i=0l; i<=n; ++i)
      rand();
    while (s>0) {
      if (s&1l==1l)
	b=static_cast<long>((static_cast<long long>(c)*
			     static_cast<long long>(b))%modulus);
      s>>=1l;
      c=static_cast<long>((static_cast<long long>(c)*
			   static_cast<long long>(c))%modulus);
    }
    a=b;
    backward();
  }
}

void TRNG::generic_MLCG::jump(long s) {
  long i, t;
  if (s<0l)
    throw error("invalid argument for TRNG::generic_MLCG::jump");
  else {
    t=a;
    for (i=0; i<s; ++i)
      a=static_cast<long>((static_cast<long long>(a)*
			   static_cast<long long>(a))%modulus);
    rand();
    a=t;
  }
}

void TRNG::generic_MLCG::save_status(vector<long> &s) {
  s.resize(4);
  s[0]=type;
  s[1]=a;
  s[2]=r;
  s[3]=modulus;
}

void TRNG::generic_MLCG::load_status(vector<long> &s) {
  if ((s.size()!=4) || (s[0]!=type))
    throw error("TRNG::generic_MLCG::load_status wrong parameter");
  else {
    a=s[1];
    r=s[2];
    modulus=s[3];
    max_val=modulus-1l;;
    max_val2=max_val/2l;
  }
}

TRNG::generic_MLCG::generic_MLCG(long a0=16807l, long m=2147483647l, 
				 long r0=1l) {
  modulus=m;
  a=a0;
  max_val=modulus-1l;
  max_val2=max_val/2l;
  r_save=r0;
  a_save=a0;
  modulus_save=m;
}


//----------------------------------------------------------------------


const char * TRNG::ParkMiller::name(void) {
  return("ParkMiller");
}

void TRNG::ParkMiller::reset(void) {
  R.reset();
  max_val=R.max()-1l;
  max_val2=max_val/2l;
}

void TRNG::ParkMiller::seed(long s=0l) {
  R.seed(s);
}

long TRNG::ParkMiller::rand(void) {
  return(R.rand()-1l);
}

RNG *TRNG::ParkMiller::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new ParkMiller);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::ParkMiller::split(long s, long n) {
  R.split(s, n);
}

void TRNG::ParkMiller::jump(long s) {
  if (s<0l)
    throw error("invalid argument for TRNG::ParkMiller::jump");
  else 
    R.jump(s);
}

void TRNG::ParkMiller::save_status(vector<long> &s) {
  R.save_status(s);
  s[0]=type;
}

void TRNG::ParkMiller::load_status(vector<long> &s) {
  if ((s.size()!=4) || (s[0]!=type))
    throw error("TRNG::ParkMiller::load_status wrong parameter");
  else {
    s[0]=generic_MLCG_t;
    R.load_status(s);
    s[0]=type;
  }
}

TRNG::ParkMiller::ParkMiller() : R(16807l, 2147483647l, 1l) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::RAND::name(void) {
  return("RAND");
}

void TRNG::RAND::backward(void) {
  long i;
  for (i=0l; i<32l; ++i)
    jump(i);
}

void TRNG::RAND::reset(void) {
  a=69069l;
  b=1l;
  r=1l;
}

void TRNG::RAND::seed(long s=0l) {
  if (s<0l)
    throw error("bad seed for TRNG::RAND::seed");
  r=s;
}

long TRNG::RAND::rand(void) {
  r=a*r+b;
  return(r/2l);
}

RNG *TRNG::RAND::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new RAND);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::RAND::split(long s, long n) {
  long i;
  unsigned long t1, t2;
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::RAND::split");
  if (s>1l) {
    t1=1ul;
    t2=0ul;
    for (i=0l; i<=n; ++i)
      rand();
    for (i=0l; i<s; ++i) {
      t2+=t1;
      t1*=a;
    }
    b*=t2;
    a=t1;
    backward();
  }
}

void TRNG::RAND::jump(long s) {
  long i;
  unsigned long t1, t2, t3;
  if (s<0l)
    throw error("invalid argument for TRNG::RAND::jump");
  else {
    t1=a;
    for (i=0l; i<s; ++i)
      t1*=t1;
    t2=1ul;
    t3=a;
    while (s>0l) {
      t2*=(1ul+t3);
      t3*=t3;
      --s;
    }
    r=r*t1+t2*b;
  }
}

void TRNG::RAND::save_status(vector<long> &s) {
  s.resize(4);
  s[0]=type;
  s[1]=a;  s[2]=b;  s[3]=r;
}

void TRNG::RAND::load_status(vector<long> &s) {
  if ((s.size()!=4) || (s[0]!=type))
    throw error("TRNG::RAND::load_status wrong parameter");
  else {
    a=s[1];
    b=s[2];
    r=s[3];
  }
}

TRNG::RAND::RAND() {
  reset();
  max_val=2147483647l;
  max_val2=max_val/2l;
}


//----------------------------------------------------------------------


void TRNG::LCG64::backward(void) {
  long i;
  for (i=0l; i<64l; ++i)
    jump(i);
}

const char * TRNG::LCG64::name(void) {
  return("LCG64");
}

void TRNG::LCG64::reset(void) {
  a=18145460002477866997ull;
  b=1ull;
  r=1ull;
}

void TRNG::LCG64::seed(long s=0l) {
  r=(static_cast<unsigned long long>(s)<<32)+1ull;
}

long TRNG::LCG64::rand(void) {
  r=a*r+b;
  return(static_cast<long>(r>>33));
}

RNG *TRNG::LCG64::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new LCG64);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::LCG64::split(long s, long n) {
  long i;
  unsigned long long t1, t2;
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::LCG64::split");
  if (s>1l) {
    t1=1ull;
    t2=0ull;
    for (i=0l; i<=n; ++i)
      rand();
    for (i=0l; i<s; ++i) {
      t2+=t1;
      t1*=a;
    }
    b*=t2;
    a=t1;
    backward();
  }
}

void TRNG::LCG64::jump(long s) {
  long i;
  unsigned long long t1, t2, t3;
  if (s<0l)
    throw error("invalid argument for TRNG::LCG64::jump");
  else {
    t1=a;
    for (i=0l; i<s; ++i)
      t1*=t1;
    t2=1ull;
    t3=a;
    while (s>0l) {
      t2*=(1ull+t3);
      t3*=t3;
      --s;
    }
    r=r*t1+t2*b;
  }
}

void TRNG::LCG64::save_status(vector<long> &s) {
  s.resize(7);
  s[0]=type;
  s[1]=static_cast<long>(a>>32);
  s[2]=static_cast<long>(a&0xffffffffull);
  s[3]=static_cast<long>(b>>32);
  s[4]=static_cast<long>(b&0xffffffffull);
  s[5]=static_cast<long>(r>>32);
  s[6]=static_cast<long>(r&0xffffffffull);
}

void TRNG::LCG64::load_status(vector<long> &s) {
  if ((s.size()!=7) || (s[0]!=type))
    throw error("TRNG::LCG64::load_status wrong parameter");
  else {
    a=(static_cast<unsigned long long>(s[1])<<32)|
      (static_cast<unsigned long long>
       (static_cast<unsigned long>(s[2])));
    b=(static_cast<unsigned long long>(s[3])<<32)|
      (static_cast<unsigned long long>
       (static_cast<unsigned long>(s[4])));
    r=(static_cast<unsigned long long>(s[5])<<32)|
      (static_cast<unsigned long long>
       (static_cast<unsigned long>(s[6])));
  }
}

TRNG::LCG64::LCG64() {
  reset();
  max_val=0x7fffffffl;
  max_val2=max_val/2l;
}


//----------------------------------------------------------------------


const char * TRNG::LMC2::name(void) {
  return("LMC2");
}

void TRNG::LMC2::backward(void) {
  long t;
  t=r[0];
  t-=static_cast<long>((static_cast<long long>(a[0])*
			static_cast<long long>(r[1]))%modulus);
  if (t<0l)
    t+=modulus;
  t=static_cast<long>((static_cast<long long>(t)*
		       static_cast<long long>
		       (modulo_invers(a[1], modulus)))%modulus);
  r[0]=r[1];  r[1]=t;
}

void TRNG::LMC2::reset(void) {
  vector<long> s(6);
  s[0]=type;
  s[1]=523007613l;  s[2]=756894486l;  // multiplier
  s[3]=0l;          s[4]=1l;          // seed
  s[5]=2147483647l;                   // modulus
  load_status(s);  
}

void TRNG::LMC2::seed(long s=0l) {
  if (s<0l) 
    throw error("invalid argument for LMC2::seed");
  else {
    r[0]=s;  r[1]=1l;
  }
}

long TRNG::LMC2::rand(void) {
  long t;
  t=static_cast<long>
    (((static_cast<long long>(a[0])*
       static_cast<long long>(r[0]))%modulus+
      (static_cast<long long>(a[1])*
       static_cast<long long>(r[1]))%modulus)
     %modulus);
  r[1]=r[0];  r[0]=t;
  return(r[0]);
}

RNG *TRNG::LMC2::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new LMC2);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::LMC2::split(long s, long n) {
  long i, q0, q1, q2, q3;
  vector<long> b(4);
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::LMC2::split");
  else 
    if (s>1l) {
      for (i=0l; i<=n; ++i) 
	q0=rand();
      for (i=0l; i<s; ++i)
	q1=rand();
      for (i=0l; i<s; ++i)
	q2=rand();
      for (i=0l; i<s; ++i)
	q3=rand();
      b[0]=q1;  b[1]=q0;  
      b[2]=q2;  b[3]=q1;
      a[0]=q2;  a[1]=q3;
      gauss(b, a, modulus);
      r[0]=q1;  r[1]=q0;
      backward();
      backward();
    }
}

void TRNG::LMC2::jump(long s) {
  long i, t1, t2;
  vector<long> b(4), c(4), d(2);
  if (s<0l)
    throw error("invalid argument for TRNG::LMC2::jump");
  else {
    t1=a[0];    t2=a[1];
    b[0]=a[0];  b[1]=a[1];
    b[2]=1l;    b[3]=0l;
    for (i=0l; i<s; ++i)
      if ((i&1l)==0l)
	matrix_mult(b, b, c, modulus);
      else 
	matrix_mult(c, c, b, modulus);
    if ((s&1l)==0l)
      matrix_vec_mult(b, r, d, modulus);
    else
      matrix_vec_mult(c, r, d, modulus);
    r[0]=d[0];  r[1]=d[1];
    a[0]=t1;    a[1]=t2;
  }
}

void TRNG::LMC2::save_status(vector<long> &s) {
  s.resize(6);
  s[0]=type;
  s[1]=a[0];  s[2]=a[1];
  s[3]=r[0];  s[4]=r[1];
  s[5]=modulus;
}

void TRNG::LMC2::load_status(vector<long> &s) {
  if ((s.size()!=6) || (s[0]!=type))
    throw error("TRNG::LMC2::load_status wrong parameter");
  else {
    a[0]=s[1];  a[1]=s[2];
    r[0]=s[3];  r[1]=s[4];
    modulus=s[5];
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::LMC2::LMC2() : a(2), r(2) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::LMC3::name(void) {
  return("LMC3");
}

void TRNG::LMC3::backward(void) {
  long t;
  t=r[0];
  t-=static_cast<long>((static_cast<long long>(a[0])*
			static_cast<long long>(r[1]))%modulus);
  if (t<0l)
    t+=modulus;
  t-=static_cast<long>((static_cast<long long>(a[1])*
			static_cast<long long>(r[2]))%modulus);
  if (t<0l)
    t+=modulus;
  t=static_cast<long>((static_cast<long long>(t)*
		       static_cast<long long>
		       (modulo_invers(a[2], modulus)))%modulus);
  r[0]=r[1];  r[1]=r[2];  r[2]=t;
}

void TRNG::LMC3::reset(void) {
  vector<long> s(8);
  s[0]=type;
  s[1]=523007613l;  s[2]=756894484l;  // multiplier
  s[3]=935294381l;
  s[4]=0l;          s[5]=1l;          // seed
  s[6]=1l;
  s[7]=2147483647l;                   // modulus
  load_status(s);  
}

void TRNG::LMC3::seed(long s=0l) {
  if (s<0l)
    throw error("invalid argument for LMC3::seed");
  else {
    r[0]=s;  r[1]=1l;  r[2]=1l;
  }
}

long TRNG::LMC3::rand(void) {
  long t;
  t=static_cast<long>
    (((static_cast<long long>(a[0])*
       static_cast<long long>(r[0]))%modulus+
      (static_cast<long long>(a[1])*
       static_cast<long long>(r[1]))%modulus+
      (static_cast<long long>(a[2])*
       static_cast<long long>(r[2]))%modulus)
     %modulus);
  r[2]=r[1];  r[1]=r[0];  r[0]=t;
  return(r[0]);
}

RNG *TRNG::LMC3::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new LMC3);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::LMC3::split(long s, long n) {
  long i, q0, q1, q2, q3, q4, q5;
  vector<long> b(9);
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::LMC3::split");
  else 
    if (s>1l) {
      for (i=0l; i<=n; ++i) 
	q0=rand();
      for (i=0l; i<s; ++i)
	q1=rand();
      for (i=0l; i<s; ++i)
	q2=rand();
      for (i=0l; i<s; ++i)
	q3=rand();
      for (i=0l; i<s; ++i)
	q4=rand();
      for (i=0l; i<s; ++i)
	q5=rand();
      b[0]=q2;  b[1]=q1;  b[2]=q0;
      b[3]=q3;  b[4]=q2;  b[5]=q1;
      b[6]=q4;  b[7]=q3;  b[8]=q2;
      a[0]=q3;  a[1]=q4;  a[2]=q5;
      gauss(b, a, modulus);
      r[0]=q2;  r[1]=q1;  r[2]=q0;
      backward();
      backward();
      backward();
    }
}

void TRNG::LMC3::jump(long s) {
  long i, t1, t2, t3;
  vector<long> b(9), c(9), d(3);
  if (s<0l)
    throw error("invalid argument for TRNG::LMC3::jump");
  else {
    t1=a[0];    t2=a[1];    t3=a[2];
    b[0]=a[0];  b[1]=a[1];  b[2]=a[2];
    b[3]=1l;    b[4]=0l;    b[5]=0l;
    b[6]=0l;    b[7]=1l;    b[8]=0l;
    for (i=0l; i<s; ++i)
      if ((i&1l)==0l)
	matrix_mult(b, b, c, modulus);
      else 
	matrix_mult(c, c, b, modulus);
    if ((s&1l)==0l)
      matrix_vec_mult(b, r, d, modulus);
    else
      matrix_vec_mult(c, r, d, modulus);
    r[0]=d[0];  r[1]=d[1];  r[2]=d[2];
    a[0]=t1;    a[1]=t2;    a[2]=t3;
  }
}

void TRNG::LMC3::save_status(vector<long> &s) {
  s.resize(8);
  s[0]=type;
  s[1]=a[0];  s[2]=a[1];  s[3]=a[2];
  s[4]=r[0];  s[5]=r[1];  s[6]=r[2];
  s[7]=modulus;
}

void TRNG::LMC3::load_status(vector<long> &s) {
  if ((s.size()!=8) || (s[0]!=type))
    throw error("TRNG::LMC3::load_status wrong parameter");
  else {
    a[0]=s[1];  a[1]=s[2];  a[2]=s[3];
    r[0]=s[4];  r[1]=s[5];  r[2]=s[6];
    modulus=s[7];
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::LMC3::LMC3() : a(3), r(3) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::LMC4::name(void) {
  return("LMC4");
}

void TRNG::LMC4::backward(void) {
  long t;
  t=r[0];
  t-=static_cast<long>((static_cast<long long>(a[0])*
			static_cast<long long>(r[1]))%modulus);
  if (t<0l)
    t+=modulus;
  t-=static_cast<long>((static_cast<long long>(a[1])*
			static_cast<long long>(r[2]))%modulus);
  if (t<0l)
    t+=modulus;
  t-=static_cast<long>((static_cast<long long>(a[2])*
			static_cast<long long>(r[3]))%modulus);
  if (t<0l)
    t+=modulus;
  t=static_cast<long>((static_cast<long long>(t)*
		       static_cast<long long>
		       (modulo_invers(a[3], modulus)))%modulus);
  r[0]=r[1];  r[1]=r[2];  r[2]=r[3];  r[3]=t;
}

void TRNG::LMC4::reset(void) {
  vector<long> s(10);
  s[0]=type;
  s[1]=523007613l;  s[2]=210014925l;  // multiplier
  s[3]=1033683247l; s[4]=935294388l;
  s[5]=0l;          s[6]=1l;          // seed
  s[7]=1l;          s[8]=1l;
  s[9]=2147483647l;                   // modulus
  load_status(s);
}

void TRNG::LMC4::seed(long s=0l) {
  if (s<0l)
    throw error("invalid argument for LMC4::seed");
  else {
    r[0]=s;  r[1]=1l;  r[2]=1l;  r[3]=1l;
  }
}

long TRNG::LMC4::rand(void) {
  long t;
  t=static_cast<long>
    (((static_cast<long long>(a[0])*
       static_cast<long long>(r[0]))%modulus+
      (static_cast<long long>(a[1])*
       static_cast<long long>(r[1]))%modulus+
      (static_cast<long long>(a[2])*
       static_cast<long long>(r[2]))%modulus+
      (static_cast<long long>(a[3])*
       static_cast<long long>(r[3]))%modulus)
     %modulus);
  r[3]=r[2];  r[2]=r[1];  r[1]=r[0];  r[0]=t;
  return(r[0]);
}

RNG *TRNG::LMC4::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new LMC4);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::LMC4::split(long s, long n) {
  long i, q0, q1, q2, q3, q4, q5, q6, q7;
  vector<long> b(16);
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::LMC4::split");
  else 
    if (s>1l) {
      for (i=0l; i<=n; ++i) 
	q0=rand();
      for (i=0l; i<s; ++i)
	q1=rand();
      for (i=0l; i<s; ++i)
	q2=rand();
      for (i=0l; i<s; ++i)
	q3=rand();
      for (i=0l; i<s; ++i)
	q4=rand();
      for (i=0l; i<s; ++i)
	q5=rand();
      for (i=0l; i<s; ++i)
	q6=rand();
      for (i=0l; i<s; ++i)
	q7=rand();
      b[ 0]=q3;  b[ 1]=q2;  b[ 2]=q1;  b[ 3]=q0;
      b[ 4]=q4;  b[ 5]=q3;  b[ 6]=q2;  b[ 7]=q1;
      b[ 8]=q5;  b[ 9]=q4;  b[10]=q3;  b[11]=q2;
      b[12]=q6;  b[13]=q5;  b[14]=q4;  b[15]=q3;
      a[ 0]=q4;  a[ 1]=q5;  a[ 2]=q6;  a[ 3]=q7;
      gauss(b, a, modulus);
      r[0]=q3;   r[1]=q2;   r[2]=q1;   r[3]=q0;
      backward();
      backward();
      backward();
      backward();
    }
}

void TRNG::LMC4::jump(long s) {
  long i, t1, t2, t3, t4;
  vector<long> b(16), c(16), d(4);
  if (s<0l)
    throw error("invalid argument for TRNG::LMC4::jump");
  else {
    t1=a[0];     t2=a[1];     t3=a[2];     t4=a[3];
    b[ 0]=a[0];  b[ 1]=a[1];  b[ 2]=a[2];  b[ 3]=a[3];
    b[ 4]=1l;    b[ 5]=0l;    b[ 6]=0l;    b[ 7]=0l;
    b[ 8]=0l;    b[ 9]=1l;    b[10]=0l;    b[11]=0l;
    b[12]=0l;    b[13]=0l;    b[14]=1l;    b[15]=0l;
    for (i=0l; i<s; ++i)
      if ((i&1l)==0l)
	matrix_mult(b, b, c, modulus);
      else 
	matrix_mult(c, c, b, modulus);
    if ((s&1l)==0l)
      matrix_vec_mult(b, r, d, modulus);
    else
      matrix_vec_mult(c, r, d, modulus);
    r[0]=d[0];  r[1]=d[1];  r[2]=d[2];  r[3]=d[3];
    a[0]=t1;    a[1]=t2;    a[2]=t3;    a[3]=t4;
  }
}

void TRNG::LMC4::save_status(vector<long> &s) {
  s.resize(10);
  s[0]=type;
  s[1]=a[0];  s[2]=a[1];  s[3]=a[2];  s[4]=a[3];
  s[5]=r[0];  s[6]=r[1];  s[7]=r[2];  s[8]=r[3]; 
  s[9]=modulus;
}

void TRNG::LMC4::load_status(vector<long> &s) {
  if ((s.size()!=10) || (s[0]!=type))
    throw error("TRNG::LMC4::load_status wrong parameter");
  else {
    a[0]=s[1];  a[1]=s[2];  a[2]=s[3];  a[3]=s[4];
    r[0]=s[5];  r[1]=s[6];  r[2]=s[7];  r[3]=s[8];
    modulus=s[9];
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::LMC4::LMC4() : a(4), r(4) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::CLCG2::name(void) {
  return("CLCG2");
}

void TRNG::CLCG2::reset(void) {
  R1.reset();
  R2.reset();
  modulus=R1.max();
  max_val=modulus-1l;
  max_val2=max_val/2l;
}

void TRNG::CLCG2::seed(long s=0l) {
  if (s<0l)
    throw error("bad seed for TRNG::CLCG2::seed");
  R1.seed(s);
  R2.seed(s);
}

long TRNG::CLCG2::rand(void) {
  return(static_cast<long>((static_cast<long long>(R1.rand())+
			    static_cast<long long>(R2.rand()))%
			   modulus));
}

RNG *TRNG::CLCG2::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new CLCG2);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::CLCG2::split(long s, long n) {
  R1.split(s, n);
  R2.split(s, n);
}

void TRNG::CLCG2::jump(long s) {
  if (s<0l)
    throw error("invalid argument for TRNG::CLCG2::jump");
  else {
    R1.jump(s);
    R2.jump(s);
  }
}

void TRNG::CLCG2::save_status(vector<long> &s) {
  vector<long> t;
  s.resize(7);
  R1.save_status(t);
  s[0]=type;
  s[1]=t[1];  s[2]=t[2];  s[3]=t[3];
  R2.save_status(t);
  s[4]=t[1];  s[5]=t[2];  s[6]=t[3];
}

void TRNG::CLCG2::load_status(vector<long> &s) {
  vector<long> t(4);
  if ((s.size()!=7) || (s[0]!=type))
    throw error("TRNG::CLCG2::load_status wrong parameter");
  else {
    t[0]=generic_MLCG_t;
    t[1]=s[1];  t[2]=s[2];  t[3]=s[3];
    R1.load_status(t);
    t[1]=s[4];  t[2]=s[5];  t[3]=s[6];
    R2.load_status(t);    
    modulus=R1.max();
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::CLCG2::CLCG2() : 
  R1(a1, m1, 1l), R2(a2, m2, 1l) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::CLCG3::name(void) {
  return("CLCG3");
}

void TRNG::CLCG3::reset(void) {
  R1.reset();
  R2.reset();
  R3.reset();
  modulus=R1.max();
  max_val=modulus-1l;
  max_val2=max_val/2l;
}

void TRNG::CLCG3::seed(long s=0l) {
  if (s<0l)
    throw error("bad seed for TRNG::CLCG3::seed");
  R1.seed(s);
  R2.seed(s);
  R3.seed(s);
}

long TRNG::CLCG3::rand(void) {
  return(static_cast<long>((static_cast<long long>(R1.rand())+
			    static_cast<long long>(R2.rand())+
			    static_cast<long long>(R3.rand()))%
			   modulus));
}

RNG *TRNG::CLCG3::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new CLCG3);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::CLCG3::split(long s, long n) {
  R1.split(s, n);
  R2.split(s, n);
  R3.split(s, n);
}

void TRNG::CLCG3::jump(long s) {
  if (s<0l)
    throw error("invalid argument for TRNG::CLCG3::jump");
  else {
    R1.jump(s);
    R2.jump(s);
    R3.jump(s);
  }
}

void TRNG::CLCG3::save_status(vector<long> &s) {
  vector<long> t;
  s.resize(10);
  R1.save_status(t);
  s[0]=type;
  s[1]=t[1];  s[2]=t[2]; s[3]=t[3];
  R2.save_status(t);
  s[4]=t[1];  s[5]=t[2]; s[6]=t[3];
  R3.save_status(t);
  s[7]=t[1];  s[8]=t[2]; s[9]=t[3];
}

void TRNG::CLCG3::load_status(vector<long> &s) {
  vector<long> t(4);
  if ((s.size()!=10) || (s[0]!=type))
    throw error("TRNG::CLCG3::load_status wrong parameter");
  else {
    t[0]=generic_MLCG_t;
    t[1]=s[1];  t[2]=s[2];  t[3]=s[3];
    R1.load_status(t);
    t[1]=s[4];  t[2]=s[5];  t[3]=s[6];
    R2.load_status(t);    
    t[1]=s[7];  t[2]=s[8];  t[3]=s[9];
    R3.load_status(t);    
    modulus=R1.max();
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::CLCG3::CLCG3() : 
  R1(a1, m1, 1l), R2(a2, m2, 1l), R3(a3, m3, 1l) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::CLCG4::name(void) {
  return("CLCG4");
}

void TRNG::CLCG4::reset(void) {
  R1.reset();
  R2.reset();
  R3.reset();
  R4.reset();
  modulus=R1.max();
  max_val=modulus-1l;
  max_val2=max_val/2l;
}

void TRNG::CLCG4::seed(long s=0l) {
  if (s<0l)
    throw error("bad seed for TRNG::CLCG4::seed");
  R1.seed(s);
  R2.seed(s);
  R3.seed(s);
  R4.seed(s);
}

long TRNG::CLCG4::rand(void) {
  return(static_cast<long>((static_cast<long long>(R1.rand())+
			    static_cast<long long>(R2.rand())+
			    static_cast<long long>(R3.rand())+
			    static_cast<long long>(R4.rand()))%
			   modulus));
}

RNG *TRNG::CLCG4::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new CLCG4);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::CLCG4::split(long s, long n) {
  R1.split(s, n);
  R2.split(s, n);
  R3.split(s, n);
  R4.split(s, n);
}

void TRNG::CLCG4::jump(long s) {
  if (s<0l)
    throw error("invalid argument for TRNG::CLCG4::jump");
  else {
    R1.jump(s);
    R2.jump(s);
    R3.jump(s);
    R4.jump(s);
  }
}

void TRNG::CLCG4::save_status(vector<long> &s) {
  vector<long> t;
  s.resize(13);
  R1.save_status(t);
  s[0]=type;
  s[1]=t[1];  s[2]=t[2];  s[3]=t[3];
  R2.save_status(t);
  s[4]=t[1];  s[5]=t[2];  s[6]=t[3];
  R3.save_status(t);
  s[7]=t[1];  s[8]=t[2];  s[9]=t[3];
  R4.save_status(t);
  s[10]=t[1];  s[11]=t[2];  s[12]=t[3];
}

void TRNG::CLCG4::load_status(vector<long> &s) {
  vector<long> t(4);
  if ((s.size()!=13) || (s[0]!=type))
    throw error("TRNG::CLCG4::load_status wrong parameter");
  else {
    t[0]=generic_MLCG_t;
    t[1]=s[1];  t[2]=s[2];  t[3]=s[3];
    R1.load_status(t);
    t[1]=s[4];  t[2]=s[5];  t[3]=s[6];
    R2.load_status(t);
    t[1]=s[7];  t[2]=s[8];  t[3]=s[9];
    R3.load_status(t);
    t[1]=s[10];  t[2]=s[11];  t[3]=s[12];
    R4.load_status(t); 
    modulus=R1.max();
    max_val=modulus-1l;
    max_val2=max_val/2l;
  }
}

TRNG::CLCG4::CLCG4() : 
  R1(a1, m1, 1l), R2(a2, m2, 1l), R3(a3, m3, 1l), R4(a4, m4, 1l) {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::EINV::name(void) {
  return("EINV");
}

long TRNG::EINV::modulo_invers(long a) {
  long long t1, t2, t3;
  // special modulus invers function for 2^30+2^28+3
  t1=t2=a;
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus; 
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus; 
  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  t2=(t2*t2)%modulus;  
  t2=(t2*t2)%modulus;  
  t3=t2;
  t3=(t3*t3)%modulus;  t3=(t3*t3)%modulus;
  return(static_cast<long>((((t1*t2)%modulus)*t3)%modulus));
}

void TRNG::EINV::reset(void) {
  i=0l;
  di=1l;
  a=1073741831l;
  modulus=1342177283l;  // 2^30+2^28+3
  max_val=modulus-1l;
  max_val2=max_val/2l;
}

void TRNG::EINV::seed(long s=0l) {
  if ((s%modulus)<=0l)
    throw error("bad seed for TRNG::EINV::seed");
  i=s%modulus;
}

long TRNG::EINV::rand(void) {
  i=static_cast<long>
    ((static_cast<long long>(i)+
      static_cast<long long>(di))%modulus);
  return(modulo_invers(static_cast<long>
		       ((static_cast<long long>(a)*
			 static_cast<long long>(i))%modulus)));
}

RNG *TRNG::EINV::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new EINV);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::EINV::split(long s, long n) {
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::EINV::split");
  if (s>1l) {
    di=static_cast<long>
      ((static_cast<long long>(s)*
	static_cast<long long>(di))%modulus);
    i=static_cast<long>
      ((static_cast<long long>(i)-
	static_cast<long long>(di)+
	static_cast<long long>(n)+1ll)%modulus);
  }  
}

void TRNG::EINV::jump(long s) {
  long long t;
  long j;
  if (s<0l)
    throw error("invalid argument for TRNG::EINV::jump");
  else {
    t=1ll;
    for (j=0l; j<s; ++j)
      t=(t+t)%modulus;
    t=t*di%modulus;
    i=static_cast<long>((t+static_cast<long long>(i))%modulus);
  }
}

void TRNG::EINV::save_status(vector<long> &s) {
  s.resize(4);
  s[0]=type;
  s[1]=a;  s[2]=i;  s[3]=di;
}

void TRNG::EINV::load_status(vector<long> &s) {
  if ((s.size()!=4) || (s[0]!=type))
    throw error("TRNG::EINV::load_status wrong parameter");
  else {
    a=s[1];  i=s[2];  di=s[3];
  }
}

TRNG::EINV::EINV() {
  reset();
}


//----------------------------------------------------------------------


const char * TRNG::EINVLCG64::name(void) {
  return("EINVLCG64");
}

void TRNG::EINVLCG64::reset(void) {
  R1.reset();
  R2.reset();
  max_val=0x7fffffff;
  max_val2=max_val/2l;
}

void TRNG::EINVLCG64::seed(long s=0l) {
  R1.seed(s);
  R2.seed();
}

long TRNG::EINVLCG64::rand(void) {
  return((R1.rand()+R2.rand())&0x7fffffffl);
}

RNG *TRNG::EINVLCG64::copy(void) {
  vector<long> s;
  RNG *R=static_cast<RNG *>(new EINVLCG64);
  save_status(s);
  R->load_status(s);
  return(R);
}

void TRNG::EINVLCG64::split(long s, long n) {
  if (s<1l || n>s || n<0l)
    throw error("invalid argument for TRNG::EINVLCG64::split");
  if (s>1l) {
    R1.split(s, n);
    R2.split(s, n);
  }
}

void TRNG::EINVLCG64::jump(long s) {
  R1.jump(s);
  R2.jump(s);
}

void TRNG::EINVLCG64::save_status(vector<long> &s) {
  vector<long> t;
  s.resize(10);
  R1.save_status(t);
  s[0]=type;
  s[1]=t[1];  s[2]=t[2];  s[3]=t[3];  
  s[4]=t[4];  s[5]=t[5];  s[6]=t[6];
  R2.save_status(t);
  s[7]=t[1];  s[8]=t[2];  s[9]=t[3];
}

void TRNG::EINVLCG64::load_status(vector<long> &s) {
  vector<long> t;
  if ((s.size()!=10) || (s[0]!=type))
    throw error("TRNG::EINVLCG64::load_status wrong parameter");
  else {
    t.resize(7);
    t[0]=LCG64_t;
    t[1]=s[1];  t[2]=s[2];  t[3]=s[3];
    t[4]=s[4];  t[5]=s[5];  t[6]=s[6];
    R1.load_status(t);
    t.resize(4);
    t[0]=EINV_t;
    t[1]=s[7];  t[2]=s[8];  t[3]=s[9];
    R2.load_status(t);    
  }
}

TRNG::EINVLCG64::EINVLCG64() {
  reset();
}
